@@ -26,7 +26,6 @@ uint32_t gssntlm_init_sec_context(uint32_t *minor_status,
     struct gssntlm_ctx *ctx;
     struct gssntlm_name *server = NULL;
     struct gssntlm_cred *cred = NULL;
-    char *computer_name = NULL;
     char *nb_computer_name = NULL;
     char *nb_domain_name = NULL;
     struct gssntlm_name *client_name = NULL;
@@ -117,6 +116,10 @@ uint32_t gssntlm_init_sec_context(uint32_t *minor_status,
         ctx->gss_flags = req_flags;
 
         ctx->neg_flags = NTLMSSP_DEFAULT_CLIENT_FLAGS;
+        /* override neg_flags default if requested */
+        if (cred->neg_flags) {
+            ctx->neg_flags = cred->neg_flags;
+        }
 
         /*
          * we ignore unsupported flags for now
@@ -165,13 +168,8 @@ uint32_t gssntlm_init_sec_context(uint32_t *minor_status,
             if (retmaj) goto done;
         }
 
-        computer_name = strdup(client_name->data.server.name);
-        if (!computer_name) {
-            set_GSSERR(ENOMEM);
-            goto done;
-        }
-
-        retmin = netbios_get_names(ctx->external_context, computer_name,
+        retmin = netbios_get_names(ctx->external_context,
+                                   client_name->data.server.name,
                                    &nb_computer_name, &nb_domain_name);
         if (retmin) {
             set_GSSERR(retmin);
@@ -398,7 +396,7 @@ uint32_t gssntlm_init_sec_context(uint32_t *minor_status,
             ctx->gss_flags |= GSS_C_INTEG_FLAG;
         }
         if (ctx->neg_flags & NTLMSSP_NEGOTIATE_SEAL) {
-            ctx->gss_flags |= GSS_C_CONF_FLAG & GSS_C_INTEG_FLAG;
+            ctx->gss_flags |= GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG;
         }
 
         ctx->stage = NTLMSSP_STAGE_DONE;
@@ -433,7 +431,6 @@ done:
         gssntlm_release_cred(&tmpmin, (gss_cred_id_t *)&cred);
     }
     gssntlm_release_name(&tmpmin, (gss_name_t *)&client_name);
-    safefree(computer_name);
     safefree(nb_computer_name);
     safefree(nb_domain_name);
     safefree(trgt_name);
@@ -532,7 +529,6 @@ uint32_t gssntlm_accept_sec_context(uint32_t *minor_status,
     int lm_compat_lvl = -1;
     struct ntlm_buffer challenge = { 0 };
     struct gssntlm_name *server_name = NULL;
-    char *computer_name = NULL;
     char *nb_computer_name = NULL;
     char *nb_domain_name = NULL;
     char *chal_target_name;
@@ -618,13 +614,8 @@ uint32_t gssntlm_accept_sec_context(uint32_t *minor_status,
             goto done;
         }
 
-        computer_name = strdup(server_name->data.server.name);
-        if (!computer_name) {
-            set_GSSERR(ENOMEM);
-            goto done;
-        }
-
-        retmin = netbios_get_names(ctx->external_context, computer_name,
+        retmin = netbios_get_names(ctx->external_context,
+                                   server_name->data.server.name,
                                    &nb_computer_name, &nb_domain_name);
         if (retmin) {
             set_GSSERR(retmin);
@@ -648,6 +639,11 @@ uint32_t gssntlm_accept_sec_context(uint32_t *minor_status,
         ctx->neg_flags = NTLMSSP_DEFAULT_SERVER_FLAGS;
         /* Fixme: How do we allow anonymous negotition ? */
 
+        /* override neg_flags default if requested */
+        if (cred && cred->neg_flags) {
+            ctx->neg_flags = cred->neg_flags;
+        }
+
         if (gssntlm_sec_lm_ok(ctx)) {
             ctx->neg_flags |= NTLMSSP_REQUEST_NON_NT_SESSION_KEY;
             ctx->neg_flags |= NTLMSSP_NEGOTIATE_LM_KEY;
@@ -686,6 +682,18 @@ uint32_t gssntlm_accept_sec_context(uint32_t *minor_status,
 
             /* leave only the crossing between requested and allowed flags */
             ctx->neg_flags &= in_flags;
+
+            /* Try to force the use of NTLMSSP_NEGOTIATE_VERSION even if the
+             * client did not advertize it in their negotiate message, but
+             * should be capable of providing it.
+             * This is what Windows Server 2022 also does, and addresses
+             * issues with older clients that incorrectly deal with MIC
+             * calculations in absence of this flag. */
+            if ((ctx->neg_flags & (NTLMSSP_NEGOTIATE_ALWAYS_SIGN |
+                                    NTLMSSP_NEGOTIATE_SEAL |
+                                    NTLMSSP_NEGOTIATE_SIGN))) {
+                ctx->neg_flags |= NTLMSSP_NEGOTIATE_VERSION;
+            }
         } else {
             /* If there is no negotiate message set datagram mode */
             ctx->neg_flags |= NTLMSSP_NEGOTIATE_DATAGRAM | \
@@ -731,15 +739,28 @@ uint32_t gssntlm_accept_sec_context(uint32_t *minor_status,
             goto done;
         }
 
+        /* TODO: allow client applications to set a context option to
+         * provide an av_flags default value so that flags like
+         * MSVAVFLAGS_UNVERIFIED_SPN can be set. By default SSPI does
+         * not set this flag, and setting it causes servers with restrictive
+         * policy to fail authentication. Given no MS client set this flag
+         * by default, neither should we until there is a clear need. We
+         * trust our calling applications to do the right thing here.
+         * av_flags = MSVAVFLAGS_UNVERIFIED_SPN;
+         */
+
         timestamp = ntlm_timestamp_now();
 
         retmin = ntlm_encode_target_info(ctx->ntlm,
                                          nb_computer_name,
                                          nb_domain_name,
-                                         computer_name,
+                                         server_name->data.server.name,
                                          NULL, NULL,
-                                         NULL, &timestamp,
-                                         NULL, NULL, NULL,
+                                         av_flags ? &av_flags : NULL, /* don't include empty MsvAvFlags */
+                                         &timestamp,
+                                         NULL,
+                                         server_name->data.server.spn,
+                                         NULL,
                                          &target_info);
         if (retmin) {
             set_GSSERR(retmin);
@@ -835,15 +856,19 @@ uint32_t gssntlm_accept_sec_context(uint32_t *minor_status,
             goto done;
         }
 
-        if (((usr_name == NULL) || (usr_name[0] == '\0')) &&
-            (nt_chal_resp.length == 0) &&
-            (((lm_chal_resp.length == 1) && (lm_chal_resp.data[0] == '\0')) ||
-             (lm_chal_resp.length == 0))) {
-            /* Anonymous auth */
-            /* FIXME: not supported for now */
-            set_GSSERR(ERR_NOTSUPPORTED);
-            goto done;
-
+        if ((usr_name == NULL) || (usr_name[0] == '\0')) {
+            if ((nt_chal_resp.length == 0) &&
+                (((lm_chal_resp.length == 1) &&
+                  (lm_chal_resp.data[0] == '\0')) ||
+                 (lm_chal_resp.length == 0))) {
+                /* Anonymous auth */
+                /* FIXME: not supported for now */
+                set_GSSERR(ERR_NOTSUPPORTED);
+                goto done;
+            } else {
+                set_GSSERR(ERR_NOUSRFOUND);
+                goto done;
+            }
         } else {
 
             char useratdom[1024];
@@ -1024,7 +1049,6 @@ done:
     gssntlm_release_name(&tmpmin, (gss_name_t *)&server_name);
     gssntlm_release_name(&tmpmin, (gss_name_t *)&gss_usrname);
     gssntlm_release_cred(&tmpmin, (gss_cred_id_t *)&usr_cred);
-    safefree(computer_name);
     safefree(nb_computer_name);
     safefree(nb_domain_name);
     safefree(usr_name);
