@@ -78,12 +78,12 @@ int ntlm_init_ctx(struct ntlm_ctx **ctx)
     _ctx = calloc(1, sizeof(struct ntlm_ctx));
     if (!_ctx) return ENOMEM;
 
-    _ctx->from_oem = iconv_open("UTF16LE", "UTF-8");
+    _ctx->from_oem = iconv_open("UTF-16LE", "UTF-8");
     if (_ctx->from_oem == (iconv_t) -1) {
         ret = errno;
     }
 
-    _ctx->to_oem = iconv_open("UTF-8", "UTF16LE");
+    _ctx->to_oem = iconv_open("UTF-8", "UTF-16LE");
     if (_ctx->to_oem == (iconv_t) -1) {
         iconv_close(_ctx->from_oem);
         ret = errno;
@@ -205,7 +205,6 @@ static int ntlm_str_convert(iconv_t cd,
     return 0;
 }
 
-
 uint8_t ntlmssp_sig[8] = {'N', 'T', 'L', 'M', 'S', 'S', 'P', 0};
 
 static void ntlm_encode_header(struct wire_msg_hdr *hdr, uint32_t msg_type)
@@ -256,6 +255,7 @@ static int ntlm_decode_oem_str(struct wire_field_hdr *str_hdr,
     str_offs = le32toh(str_hdr->offset);
     if ((str_offs < payload_offs) ||
         (str_offs > buffer->length) ||
+        (UINT32_MAX - str_offs < str_len) ||
         (str_offs + str_len > buffer->length)) {
         return ERR_DECODE;
     }
@@ -299,7 +299,7 @@ static int ntlm_decode_u16l_str_hdr(struct ntlm_ctx *ctx,
     char *in, *out = NULL;
     uint16_t str_len;
     uint32_t str_offs;
-    size_t outlen;
+    size_t outlen = 0;
     int ret = 0;
 
     str_len = le16toh(str_hdr->len);
@@ -308,6 +308,7 @@ static int ntlm_decode_u16l_str_hdr(struct ntlm_ctx *ctx,
     str_offs = le32toh(str_hdr->offset);
     if ((str_offs < payload_offs) ||
         (str_offs > buffer->length) ||
+        (UINT32_MAX - str_offs < str_len) ||
         (str_offs + str_len > buffer->length)) {
         return ERR_DECODE;
     }
@@ -319,13 +320,16 @@ static int ntlm_decode_u16l_str_hdr(struct ntlm_ctx *ctx,
 
     ret = ntlm_str_convert(ctx->to_oem, in, out, str_len, &outlen);
 
-    /* make sure to terminate output string */
-    out[outlen] = '\0';
-
 done:
     if (ret) {
         safefree(out);
+    } else {
+        /* make sure to terminate output string */
+        if (out) {
+            out[outlen] = '\0';
+        }
     }
+
     *str = out;
     return ret;
 }
@@ -349,15 +353,14 @@ void ntlm_internal_set_version(uint8_t major, uint8_t minor,
 
 static int ntlm_encode_version(struct ntlm_ctx *ctx,
                                struct ntlm_buffer *buffer,
-                               size_t *data_offs)
+                               size_t data_offs)
 {
-    if (*data_offs + sizeof(struct wire_version) > buffer->length) {
+    if (data_offs + sizeof(struct wire_version) > buffer->length) {
         return ERR_ENCODE;
     }
 
-    memcpy(&buffer->data[*data_offs], &ntlmssp_version,
+    memcpy(&buffer->data[data_offs], &ntlmssp_version,
            sizeof(struct wire_version));
-    *data_offs += sizeof(struct wire_version);
     return 0;
 }
 
@@ -394,6 +397,7 @@ static int ntlm_decode_field(struct wire_field_hdr *hdr,
     offs = le32toh(hdr->offset);
     if ((offs < payload_offs) ||
         (offs > buffer->length) ||
+        (UINT32_MAX - offs < len) ||
         (offs + len > buffer->length)) {
         return ERR_DECODE;
     }
@@ -683,11 +687,19 @@ int ntlm_decode_target_info(struct ntlm_ctx *ctx, struct ntlm_buffer *buffer,
             break;
         case MSV_AV_TIMESTAMP:
             if (!av_timestamp) continue;
+            if (av_len < sizeof(timestamp)) {
+                ret = ERR_DECODE;
+                goto done;
+            }
             memcpy(&timestamp, av_pair->value, sizeof(timestamp));
             timestamp = le64toh(timestamp);
             break;
         case MSV_AV_FLAGS:
             if (!av_flags) continue;
+            if (av_len < sizeof(flags)) {
+                ret = ERR_DECODE;
+                goto done;
+            }
             memcpy(&flags, av_pair->value, sizeof(flags));
             flags = le32toh(flags);
             break;
@@ -729,8 +741,6 @@ int ntlm_decode_target_info(struct ntlm_ctx *ctx, struct ntlm_buffer *buffer,
 
 done:
     if (ret) {
-        ntlm_free_buffer_data(&sh);
-        ntlm_free_buffer_data(&cb);
         safefree(nb_computer);
         safefree(nb_domain);
         safefree(dns_computer);
@@ -754,7 +764,7 @@ done:
 
 int ntlm_process_target_info(struct ntlm_ctx *ctx, bool protect,
                              struct ntlm_buffer *in,
-                             const char *server,
+                             const char *spn,
                              struct ntlm_buffer *unhashed_cb,
                              struct ntlm_buffer *out,
                              uint64_t *out_srv_time,
@@ -786,8 +796,9 @@ int ntlm_process_target_info(struct ntlm_ctx *ctx, bool protect,
         goto done;
     }
 
-    if (server && av_target_name) {
-        if (strcasecmp(server, av_target_name) != 0) {
+    if (spn && av_target_name &&
+        ((av_flags & MSVAVFLAGS_UNVERIFIED_SPN) == 0)) {
+        if (strcasecmp(spn, av_target_name) != 0) {
             ret = EINVAL;
             goto done;
         }
@@ -808,15 +819,13 @@ int ntlm_process_target_info(struct ntlm_ctx *ctx, bool protect,
         if (ret) goto done;
     }
 
-    if (!av_target_name && server) {
-        av_target_name = strdup(server);
+    if (!av_target_name && spn) {
+        av_target_name = strdup(spn);
         if (!av_target_name) {
             ret = ENOMEM;
             goto done;
         }
     }
-    /* TODO: add way to tell if the target name is verified o not,
-     * if not set av_flags |= MSVAVFLAGS_UNVERIFIED_SPN; */
 
     ret = ntlm_encode_target_info(ctx,
                                   nb_computer_name, nb_domain_name,
@@ -932,6 +941,12 @@ int ntlm_encode_neg_msg(struct ntlm_ctx *ctx, uint32_t flags,
         if (ret) goto done;
     }
 
+    if (flags & NTLMSSP_NEGOTIATE_VERSION) {
+        ret = ntlm_encode_version(ctx, &buffer,
+                                  (char *)&msg->version - (char *)msg);
+        if (ret) goto done;
+    }
+
 done:
     if (ret) {
         safefree(buffer.data);
@@ -959,6 +974,14 @@ int ntlm_decode_neg_msg(struct ntlm_ctx *ctx,
 
     neg_flags = le32toh(msg->neg_flags);
 
+    if ((neg_flags & NTLMSSP_NEGOTIATE_VERSION) == 0) {
+        /* adjust the payload offset to point to the
+         * version field, for compatibility with older
+         * clients that completely omitted the structure
+         * on the wire */
+        payload_offs -= sizeof(struct wire_version);
+    }
+
     if (domain &&
         (neg_flags & NTLMSSP_NEGOTIATE_OEM_DOMAIN_SUPPLIED)) {
         ret = ntlm_decode_oem_str(&msg->domain_name, buffer,
@@ -1003,10 +1026,6 @@ int ntlm_encode_chal_msg(struct ntlm_ctx *ctx,
 
     buffer.length = sizeof(struct wire_chal_msg);
 
-    if (flags & NTLMSSP_NEGOTIATE_VERSION) {
-        buffer.length += sizeof(struct wire_version);
-    }
-
     if ((flags & NTLMSSP_TARGET_TYPE_SERVER)
         || (flags & NTLMSSP_TARGET_TYPE_DOMAIN)) {
         if (!target_name) return EINVAL;
@@ -1035,7 +1054,8 @@ int ntlm_encode_chal_msg(struct ntlm_ctx *ctx,
 
     /* this must be first as it pushes the payload further down */
     if (flags & NTLMSSP_NEGOTIATE_VERSION) {
-        ret = ntlm_encode_version(ctx, &buffer, &data_offs);
+        ret = ntlm_encode_version(ctx, &buffer,
+                                  (char *)&msg->version - (char *)msg);
         if (ret) goto done;
     }
 
@@ -1077,6 +1097,7 @@ int ntlm_decode_chal_msg(struct ntlm_ctx *ctx,
 {
     struct wire_chal_msg *msg;
     size_t payload_offs;
+    size_t base_chal_size;
     uint32_t flags;
     char *trg = NULL;
     int ret = 0;
@@ -1089,6 +1110,16 @@ int ntlm_decode_chal_msg(struct ntlm_ctx *ctx,
     payload_offs = (char *)msg->payload - (char *)msg;
 
     flags = le32toh(msg->neg_flags);
+    base_chal_size = sizeof(struct wire_chal_msg);
+
+    if ((flags & NTLMSSP_NEGOTIATE_VERSION) == 0) {
+        /* adjust the payload offset to point to the
+         * version field, for compatibility with older
+         * clients that completely omitted the structure
+         * on the wire */
+        payload_offs -= sizeof(struct wire_version);
+        base_chal_size -= sizeof(struct wire_version);
+    }
 
     if ((flags & NTLMSSP_TARGET_TYPE_SERVER)
         || (flags & NTLMSSP_TARGET_TYPE_DOMAIN)) {
@@ -1107,7 +1138,7 @@ int ntlm_decode_chal_msg(struct ntlm_ctx *ctx,
 
     /* if we allowed a broken short challenge message from an old
      * server we must stop here */
-    if (buffer->length < sizeof(struct wire_chal_msg)) {
+    if (buffer->length < base_chal_size) {
         if (flags & NTLMSSP_NEGOTIATE_TARGET_INFO) {
             ret = ERR_DECODE;
         }
@@ -1190,9 +1221,6 @@ int ntlm_encode_auth_msg(struct ntlm_ctx *ctx,
     if (enc_sess_key) {
         buffer.length += enc_sess_key->length;
     }
-    if (flags & NTLMSSP_NEGOTIATE_VERSION) {
-        buffer.length += sizeof(struct wire_version);
-    }
     if (mic) {
         buffer.length += 16;
     }
@@ -1207,11 +1235,12 @@ int ntlm_encode_auth_msg(struct ntlm_ctx *ctx,
 
     /* this must be first as it pushes the payload further down */
     if (flags & NTLMSSP_NEGOTIATE_VERSION) {
-        ret = ntlm_encode_version(ctx, &buffer, &data_offs);
+        ret = ntlm_encode_version(ctx, &buffer,
+                                  (char *)&msg->version - (char *)msg);
         if (ret) goto done;
     }
 
-    /* this must be second as it pushes the payload further down */
+    /* this pushes the payload further down */
     if (mic) {
         memset(&buffer.data[data_offs], 0, mic->length);
         /* return the actual pointer back in the mic, as it will
@@ -1293,6 +1322,7 @@ int ntlm_decode_auth_msg(struct ntlm_ctx *ctx,
                          struct ntlm_buffer *mic)
 {
     struct wire_auth_msg *msg;
+    uint32_t neg_flags;
     size_t payload_offs;
     char *dom = NULL;
     char *usr = NULL;
@@ -1308,10 +1338,13 @@ int ntlm_decode_auth_msg(struct ntlm_ctx *ctx,
     msg = (struct wire_auth_msg *)buffer->data;
     payload_offs = (char *)msg->payload - (char *)msg;
 
-    /* this must be first as it pushes the payload further down */
-    if (flags & NTLMSSP_NEGOTIATE_VERSION) {
-        /* skip version for now */
-        payload_offs += sizeof(struct wire_version);
+    neg_flags = le32toh(msg->neg_flags);
+    if ((neg_flags & NTLMSSP_NEGOTIATE_VERSION) == 0) {
+        /* adjust the payload offset to point to the
+         * version field, for compatibility with older
+         * clients that completely omitted the structure
+         * on the wire */
+        payload_offs -= sizeof(struct wire_version);
     }
 
     /* Unconditionally copy 16 bytes for the MIC, if it was really
@@ -1319,12 +1352,24 @@ int ntlm_decode_auth_msg(struct ntlm_ctx *ctx,
      * in the NT Response, that will be fully decoded later by the caller
      * and the MIC checked otherwise these 16 bytes will just be ignored */
     if (mic) {
+        size_t mic_offs = payload_offs;
+
         if (mic->length < 16) return ERR_DECODE;
-        /* mic is at payload_offs right now */
-        if (buffer->length - payload_offs < 16) return ERR_DECODE;
-        memcpy(mic->data, &buffer->data[payload_offs], 16);
-        /* NOTE: we do not push down the payload because we do not know that
-         * the MIC is actually present yet for real */
+
+        if ((neg_flags & NTLMSSP_NEGOTIATE_VERSION) == 0) {
+            struct wire_version zver = {0};
+            /* mic is at payload_offs right now, but this offset may be
+             * wrongly reduced for compatibility with older clients,
+             * if all bytes are zeroed, then it means there was an actual
+             * empty version struct */
+            if (memcmp(&msg->version, &zver,
+                       sizeof(struct wire_version)) == 0) {
+                mic_offs += sizeof(struct wire_version);
+            }
+        }
+
+        if (buffer->length - mic_offs < 16) return ERR_DECODE;
+        memcpy(mic->data, &buffer->data[mic_offs], 16);
     }
 
     if (msg->lm_chalresp.len != 0 && lm_chalresp) {
