@@ -50,7 +50,7 @@ int NTOWFv1(const char *password, struct ntlm_key *result)
     int ret;
 
     len = strlen(password);
-    retstr = u8_conv_to_encoding("UTF16LE", iconveh_error,
+    retstr = u8_conv_to_encoding("UTF-16LE", iconveh_error,
                                  (const uint8_t *)password, len,
                                  NULL, NULL, &out);
     if (!retstr) return ERR_CRYPTO;
@@ -254,7 +254,7 @@ int NTOWFv2(struct ntlm_ctx *ctx, struct ntlm_key *nt_hash,
         offs += len;
     }
 
-    retstr = (uint8_t *)u8_conv_to_encoding("UTF16LE", iconveh_error,
+    retstr = (uint8_t *)u8_conv_to_encoding("UTF-16LE", iconveh_error,
                                             upcased, offs, NULL, NULL, &out);
     if (!retstr) return ERR_CRYPTO;
 
@@ -860,9 +860,8 @@ int ntlm_seal(uint32_t flags,
 
     h = &state->send;
 
-    if (!(flags & NTLMSSP_NEGOTIATE_SEAL) ||
-        (h->seal_handle == NULL)) {
-        return ENOTSUP;
+    if (h->seal_handle == NULL) {
+        return EINVAL;
     }
 
     ret = RC4_UPDATE(h->seal_handle, message, output);
@@ -902,9 +901,8 @@ int ntlm_unseal(uint32_t flags,
         h = &state->recv;
     }
 
-    if (!(flags & NTLMSSP_NEGOTIATE_SEAL) ||
-        (h->seal_handle == NULL)) {
-        return ENOTSUP;
+    if (h->seal_handle == NULL) {
+        return EINVAL;
     }
 
     ret = RC4_UPDATE(h->seal_handle, message, output);
@@ -972,9 +970,17 @@ int ntlm_verify_mic(struct ntlm_key *key,
 
     /* flags must be checked as they may push the payload further down */
     flags = le32toh(msg->neg_flags);
-    if (flags & NTLMSSP_NEGOTIATE_VERSION) {
-        /* skip version for now */
-        payload_offs += sizeof(struct wire_version);
+    if ((flags & NTLMSSP_NEGOTIATE_VERSION) == 0) {
+        struct wire_version zver = {0};
+        /* mic is at payload_offs right now, but this offset may
+         * need to be reduced if the sender completely omitted
+         * the version struct, as some older clients do */
+        if (memcmp(&msg->version, &zver,
+                   sizeof(struct wire_version)) != 0) {
+            /* version struct is not all zeros, this indicates the actual
+             * struct was omitted and payload is shifted down */
+            payload_offs -= sizeof(struct wire_version);
+        }
     }
 
     if (payload_offs + NTLM_SIGNATURE_SIZE > authenticate_message->length) {
